import { useEffect, useState } from 'react';
import useCallbackRef from './useCallbackRef';
import { isBoolean, trimAndFormatFilterStr } from '../utils';
import { EMPTY_ARRAY, FilterMatchEnum, FunctionDefaults } from '../constants';
/**
 * Parse options to array of MenuOptions and perform filtering (if applicable).
 * Set menuOptions state (ensure array returned).
 */
const useMenuOptions = (options, debouncedInputValue, filterMatchFrom, selectedOption, getOptionValue, getOptionLabel, getIsOptionDisabled, getFilterOptionString, filterIgnoreCase = false, filterIgnoreAccents = false, isMulti = false, async = false, hideSelectedOptions) => {
    const [menuOptions, setMenuOptions] = useState(EMPTY_ARRAY);
    const getFilterOptionStringRef = useCallbackRef(getFilterOptionString || FunctionDefaults.OPTION_FILTER);
    const getIsOptionDisabledRef = useCallbackRef(getIsOptionDisabled || FunctionDefaults.OPTION_IS_DISABLED);
    // Prevent effect from executing on search input mutations in 'async' mode (also prevents filtering from executing)
    const searchValue = !async ? debouncedInputValue : '';
    const isFilterMatchAny = filterMatchFrom === FilterMatchEnum.ANY;
    const hideSelectedOptionsOrDefault = isBoolean(hideSelectedOptions) ? hideSelectedOptions : isMulti;
    useEffect(() => {
        const selectedHash = selectedOption.length ? new Set(selectedOption.map((x) => x.value)) : null;
        const normalizedSearch = trimAndFormatFilterStr(searchValue, filterIgnoreCase, filterIgnoreAccents);
        const isOptionFilterMatch = (option) => {
            if (!normalizedSearch)
                return true;
            const filterVal = getFilterOptionStringRef(option);
            const normalizedOptionLabel = trimAndFormatFilterStr(filterVal, filterIgnoreCase, filterIgnoreAccents);
            return isFilterMatchAny
                ? normalizedOptionLabel.includes(normalizedSearch)
                : normalizedOptionLabel.substr(0, normalizedSearch.length) === normalizedSearch;
        };
        const parseMenuOption = (data) => {
            var _a;
            const value = getOptionValue(data);
            const label = getOptionLabel(data);
            const isDisabled = getIsOptionDisabledRef(data);
            const isSelected = (_a = selectedHash === null || selectedHash === void 0 ? void 0 : selectedHash.has(value)) !== null && _a !== void 0 ? _a : false;
            const menuOption = { data, value, label, isDisabled, isSelected };
            if (!isOptionFilterMatch(menuOption) || (hideSelectedOptionsOrDefault && isSelected)) {
                return;
            }
            return menuOption;
        };
        const nextMenuOptions = options.reduce((acc, option) => {
            const menuOption = parseMenuOption(option);
            menuOption && acc.push(menuOption);
            return acc;
        }, []);
        setMenuOptions(nextMenuOptions);
    }, [
        options,
        searchValue,
        getOptionValue,
        getOptionLabel,
        selectedOption,
        isFilterMatchAny,
        filterIgnoreCase,
        filterIgnoreAccents,
        getIsOptionDisabledRef,
        getFilterOptionStringRef,
        hideSelectedOptionsOrDefault
    ]);
    return menuOptions;
};
export default useMenuOptions;
